@@ -14,6 +14,7 @@ from sklearn.neural_network import MLPClassifier
 from sklearn.linear_model import SGDClassifier
 import xgboost
 from catboost import CatBoostClassifier
+import lightgbm as lgb
 
 # ML Regressors
 from sklearn.ensemble import RandomForestRegressor
@@ -53,22 +54,28 @@ warnings.filterwarnings(action='ignore', category=DataConversionWarning)
 from sklearn.exceptions import ConvergenceWarning
 ConvergenceWarning('ignore')
 
+# Explainable AI
+import lime
+import lime.lime_tabular
+import shap
+
 # Other
 import pandas as pd
 import numpy as np
 from tqdm import tqdm
+import sweetviz as sv
 
 def null(DataFrame):
     """
-    The null method gives you insight to number of null values
+    The null function gives you insight to number of null values
     """
     print(DataFrame.isnull().sum())
     plt.figure(figsize=(10,6)) #Change size if you have many features
-    sns.heatmap(DataFrame.isna(), cbar=False, cmap='viridis', yticklabels=False)
+    sns.heatmap(DataFrame.isna(), cbar=False, cmap='binary', yticklabels=False)
 
 def correlation(DataFrame):
     """
-    The correlation method gives you insight to the correlation between features
+    The correlation function gives you insight to the correlation between numerical features
     """
     corr = DataFrame.corr()
 
@@ -99,7 +106,7 @@ def FID(DataFrame, ML_model, plot=True, length=5, height=5):
     
     return df
 
-def compare_models(X_train, y_train, random_seed=None, classification=True):
+def compare_models(X, y, random_seed=None, classification=True):
     """
     Function for comparing feature importance of different ML models
     Only models with a feature_importances_ attribute are supported
@@ -113,6 +120,7 @@ def compare_models(X_train, y_train, random_seed=None, classification=True):
         dt = DecisionTreeClassifier(random_state=random_seed)
         xgb = xgboost.XGBClassifier(n_jobs=-1, verbosity = 0, use_label_encoder=False, random_state=random_seed)
         cb = CatBoostClassifier(random_state=random_seed)
+        lgb = lgb.LGBMClassifier(n_jobs=-1, random_state=random_seed)
     else:
         # Import Regression models
         rf = RandomForestRegressor(n_jobs=-1, random_state=random_seed)
@@ -121,14 +129,16 @@ def compare_models(X_train, y_train, random_seed=None, classification=True):
         dt = DecisionTreeRegressor(random_state=random_seed)
         xgb = xgboost.XGBRegressor(n_jobs=-1, verbosity = 0, random_state=random_seed)
         cb = CatBoostRegressor(random_state=random_seed)
+        lgb = lgb.LGBMRegressor(n_jobs=-1, random_state=random_seed)
 
     # Fit the models
-    rf.fit(X_train, y_train)
-    gbdt.fit(X_train, y_train)
-    ab.fit(X_train, y_train)
-    dt.fit(X_train, y_train)
-    xgb.fit(X_train, y_train)
-    cb.fit(X_train, y_train)
+    rf.fit(X, y)
+    gbdt.fit(X, y)
+    ab.fit(X, y)
+    dt.fit(X, y)
+    xgb.fit(X, y)
+    cb.fit(X, y)
+    lgb.fit(X, y)
 
     # Add the results
     model_list = {
@@ -137,11 +147,12 @@ def compare_models(X_train, y_train, random_seed=None, classification=True):
         'AdaBoost': ab,
         'Decision Tree': dt,
         'XGBoost': xgb,
-        'CatBoost': cb
+        'CatBoost': cb,
+        'LightGBM': lgb
     }
 
     # Create a dataframe to store the results of the models
-    column_names    = X_train.columns
+    column_names    = X.columns
     features = {}
     
     for key, value in model_list.items():
@@ -182,6 +193,7 @@ def automl(X_test, X_train, y_train, y_test, random_seed=None, classification=Tr
         mlp.out_activation_ = 'logistic' #used for binary classification
         xgb = xgboost.XGBClassifier(n_jobs=-1, verbosity = 0, use_label_encoder=False, random_state=random_seed)
         cb = CatBoostClassifier(random_state=random_seed)
+        lgb = lgb.LGBMClassifier(n_jobs=-1, random_state=random_seed)
     else:
         # Import ML models
         rf = RandomForestRegressor(n_jobs=-1, random_state=random_seed)
@@ -193,6 +205,7 @@ def automl(X_test, X_train, y_train, y_test, random_seed=None, classification=Tr
         mlp = MLPRegressor(random_state=random_seed)
         xgb = xgboost.XGBRegressor(n_jobs=-1, verbosity = 0, random_state=random_seed)
         cb = CatBoostRegressor(random_state=random_seed)
+        lgb = lgb.LGBMRegressor(n_jobs=-1, random_state=random_seed)
 
     # Fit the models
     rf.fit(X_train, y_train)
@@ -204,6 +217,7 @@ def automl(X_test, X_train, y_train, y_test, random_seed=None, classification=Tr
     mlp.fit(X_train, y_train)
     xgb.fit(X_train, y_train)
     cb.fit(X_train, y_train)
+    lgb.fit(X_train, y_train)
 
     # Add the results
     model_list = {
@@ -215,7 +229,8 @@ def automl(X_test, X_train, y_train, y_test, random_seed=None, classification=Tr
         'SVM': svm,
         'MLP': mlp,
         'XGBoost': xgb,
-        'CatBoost': cb
+        'CatBoost': cb,
+        'LightGBM': lgb
     }
 
     for key, value in model_list.items():
@@ -300,11 +315,10 @@ def automl(X_test, X_train, y_train, y_test, random_seed=None, classification=Tr
 
     return df_sorted.style.highlight_max(axis=0)
 
-# Function for hyperparameter tuning for models using GridSearchCV
-def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_score, random_seed=None, classification=True):
+def tuning(X, y, model='rf', GridSearch=False, scoring=accuracy_score, random_seed=None, classification=True):
     """
     Function for hyperparameter tuning for models using GridSearchCV or RandomizedSearchCV (default)
-    Model: 'rf', 'gbdt', 'ab', 'dt', 'knn', 'svm', 'mlp', 'xgb' or 'cb'
+    Model: 'rf', 'gbdt', 'ab', 'dt', 'knn', 'svm', 'mlp', 'xgb', 'cb', 'lgb'
     scoring: accuracy_score, precision_score, recall_score, f1_score, mean_squared_error, r2_score
     """
 
@@ -320,6 +334,7 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
         mlp.out_activation_ = 'logistic' #used for binary classification
         xgb = xgboost.XGBClassifier(n_jobs=-1, verbosity = 0, use_label_encoder=False, random_state=random_seed)
         cb = CatBoostClassifier(random_state=random_seed)
+        lgb = lgb.LGBMClassifier(n_jobs=-1, random_state=random_seed)
     else:
         # Import ML models
         rf = RandomForestRegressor(n_jobs=-1, random_state=random_seed)
@@ -331,6 +346,7 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
         mlp = MLPRegressor(random_state=random_seed)
         xgb = xgboost.XGBRegressor(n_jobs=-1, verbosity = 0, random_state=random_seed)
         cb = CatBoostRegressor(random_state=random_seed)
+        lgb = lgb.LGBMRegressor(n_jobs=-1, random_state=random_seed)
 
     # The models to be tuned
     model_list = {
@@ -342,7 +358,8 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
         'svm': svm,
         'mlp': mlp,
         'xgb': xgb,
-        'cb': cb
+        'cb': cb,
+        'lgb': lgb
     }
 
     # Hyperparameters to be tuned
@@ -396,6 +413,12 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
             'learning_rate': [0.1, 0.5, 1],
             'depth': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
             'l2_leaf_reg': [1, 5, 10, 50, 100, 500, 1000]
+        },
+        'lgb': {
+            'n_estimators': [10, 50, 100, 200, 500],
+            'learning_rate': [0.1, 0.5, 1],
+            'max_depth': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
+            'num_leaves': [2, 5, 10, 20, 50, 100, 200, 500]
         }
     }
 
@@ -405,7 +428,7 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
     if GridSearch==True:
         # GridSearchCV
         grid_search = GridSearchCV(model_list[model], param_grid=hyperparameters[model], cv=5, scoring=scorer, n_jobs=-1)
-        grid_result = grid_search.fit(X_train, y_train)
+        grid_result = grid_search.fit(X, y)
 
         print('The GridSearchCV found the following best parameters:')
         print('Best score for {}: {}'.format(model, grid_result.best_score_))
@@ -414,10 +437,54 @@ def tuning(X_train, y_train, model='rf', GridSearch=False, scoring=accuracy_scor
     else:
         # RandomizedSearchCV
         grid_search = RandomizedSearchCV(model_list[model], param_distributions=hyperparameters[model], cv=5, scoring=scorer, n_jobs=-1)
-        grid_result = grid_search.fit(X_train, y_train)
+        grid_result = grid_search.fit(X, y)
 
         print('The RandomizedSearchCV found the following best parameters:')
         print('Best score for {}: {}'.format(model, grid_result.best_score_))
         print('Best estimator for {}: {}'.format(model, grid_result.best_estimator_))
     
-    return grid_result
\ No newline at end of file
+    return grid_result
+
+def xai(X_train, X_test, model=None, mode='classification', observation=0, feature_selection='none'):
+    """
+    The eXplainable AI (XAI) function is used to explain the model's predictions. The model uses LIME to explain the predictions of the model.
+    """
+
+    if mode == 'classification':
+        explainer = lime.lime_tabular.LimeTabularExplainer(
+        training_data = X_train.values,
+        mode = mode,
+        feature_names = X_train.columns,
+        feature_selection = feature_selection #Set to 'auto' or 'lasso_path' in case of feature selection, I have set it to none to see all features
+        )
+    elif mode == 'regression':
+        explainer = lime.lime_tabular.LimeTabularExplainer(
+        training_data = X_train.values,
+        mode = mode,
+        feature_names = X_train.columns,
+        feature_selection = feature_selection #Set to 'auto' or 'lasso_path' in case of feature selection, I have set it to none to see all features
+        )
+
+    exp = explainer.explain_instance(
+        data_row=X_test.iloc[observation], #Specific observation to predict
+        predict_fn=model.predict_proba #The model GBDT in this case
+    )
+    exp.show_in_notebook(show_table=True)
+
+    exp.as_pyplot_figure();
+
+    return exp
+
+def eda(DataFrame):
+    """
+    The Exploratory Data Analysis (EDA) function is used to explore the data.
+    Using the sweetviz library, the data is visualized in a variety of ways in an HTML file.
+    """
+    
+    #EDA using Autoviz
+    sweet_report = sv.analyze([DataFrame, "DataFrame"]) #Or just type the name of the dataframe
+
+    #Saving results to HTML file
+    sweet_report.show_html('sweet_report.html')
+
+    return
\ No newline at end of file
